/*!
 * gulpis
 *
 * Copyright (c) 2016 Katsunori Koyanagi.
 * Licensed under the MIT license.
 */

"use strict";

var child_process = require("child_process");
var path = require("path");
var fs = require("fs");
var util = require("./util");
var options = require("./options");

//インストール
var install = (function () {
	//YARNがあるか
	var result = child_process.spawnSync(util.cmd("yarn"));

	//なければNPMを使用
	if (result.error) {
		return function (names) {
			names.unshift("install");
			names.unshift("--save-dev");
			names.unshift("--save-exact");

			child_process.spawnSync(util.cmd("npm"), names, { stdio: "inherit" });
		};
	}

	//YARNを使用
	return function (names) {
		names.unshift("add");
		names.push("--dev");
		names.unshift("--exact");

		child_process.spawnSync(util.cmd("yarn"), names, { stdio: "inherit" });
	};
})();

//キャッシュ
var moduleCache = {};

//通常時の処理
module.exports = {

	//起動ファイルも監視対象にする
	watchBootstrap() {
		this.loadObject(path.join(util.cwd, "gulpfile.js"));
	},

	//requireを絶対パスで行い監視する
	loadObject(pathname) {
		//絶対パスへ変換
		if (!path.isAbsolute(pathname)) {
			pathname = path.join(util.cwd, pathname);
		}
		pathname = require.resolve(pathname);

		//変化したら終了
		var watcher = fs.watch(pathname, { persistent: false });
		watcher.on("change", function () {
			if (options["internal--rerun--mode"] == "true") {
				var gulpUtil = require("gulp-util");
				gulpUtil.log("file changed.(" + pathname + ")");
				setTimeout(function () {
					gulpUtil.log("shudwon now...");
					util.restart();
				}, 100);
			}
		});

		return require(pathname);
	},

	//モジュール情報取得
	getModuleInfo(module) {
		var version;
		var path;
		var name;

		if (util.isString(module)) {
			let index = module.indexOf("@");
			if (index > 1) {
				name = module.substr(0, index);
				path = null;
				version = module.substr(index + 1);
			} else {
				name = module;
				path = null;
				version = null;
			}
		} else {
			path = module.path;
			name = module.name;
			version = module.version;
		}

		if (version) {
			if (version.indexOf("@") == -1) {
				version = "@" + version;
			}
		} else {
			version = "";
		}

		if (!path) {
			path = "";
		}

		return {
			name,
			path,
			version,
		};
	},

	//パスから検索してモジュールの存在チェック
	fastCheck(name) {
		for (let pathname of module.paths) {
			pathname = path.join(pathname, name);
			try {
				fs.statSync(pathname);
				return true;
			} catch (e) {
			}
		}

		return false;
	},

	//モジュールをセット
	writeModule(dests, key, load, name) {
		if (util.isFunction(key)) {
			key(load ? require(name) : undefined, dests);
		} else {
			dests[key] = load ? require(name) : undefined;
		}
	},

	//モジュールをまとめてロード
	loadModuleSet(requests, dests) {
		var installRequests = [];
		var installNames = [];

		//リクエスト単位で処理
		for (let request of requests) {
			let info = this.getModuleInfo(request.module);
			let load = request.load !== false;

			//インストールチェック済みならインストール処理をしない
			if (info.name in moduleCache) {
				continue;
			}

			//ディレクトリをチェックしてモジュールがあればインストール処理をしない
			if (this.fastCheck(info.name)) {
				moduleCache[info.name] = true;
				continue;
			}

			//インストール対象にする
			installRequests.push(request);
			installNames[installNames.length] = (info.path || info.name) + (info.path ? "" : info.version);
		}

		//インストールする
		if (installNames.length) {
			install(installNames);

			//成功に関係なく処理
			for (let request of installRequests) {
				let info = this.getModuleInfo(request.module);
				let load = request.load !== false;

				moduleCache[info.name] = true;
			}
		}

		//モジュールを出力
		for (let request of requests) {
			let info = this.getModuleInfo(request.module);
			let load = request.load !== false;

			this.writeModule(dests, request.key, load, info.name);
		}
	},

	//モジュールのロード
	loadModule(module, load) {
		//セットを作成してロード
		var dest = {};
		this.loadModuleSet([{
			module,
			load,
			key: "auto",
		}], dest);

		return dest.auto;
	},
};
