/*!
 * gulpis
 *
 * Copyright (c) 2016 Katsunori Koyanagi.
 * Licensed under the MIT license.
 */

"use strict";

var gulp = require("gulp");
var fs = require("fs");
var path = require("path");
var config = require("./config");
var loader = require("./loader");
var util = require("./util");
var Task = require("./task");

//タスククラスを継承しているか
function isTaskSubClass(target) {
	while (target) {
		if (target === Task) {
			return true;
		}

		target = Object.getPrototypeOf(target);
	}

	return false;
}

//オブジェクトからクラスを生成する
function createTaskClass(TaskObject) {
	if (util.isFunction(TaskObject)) {
		return createTaskClassFromFunction(TaskObject);
	}
	if (util.isArray(TaskObject)) {
		return createTaskClassFromArray(TaskObject);
	}
	return null;
}

//関数からタスククラスを生成
function createTaskClassFromFunction(TaskFunction) {
	var main;

	return class extends Task {

		configure(config) {
			main = TaskFunction.call(this, config);
		}

		execute(done) {
			return main.call(this, done);
		}
	}
}

//配列からタスククラスを生成
function createTaskClassFromArray(taskArray) {
	return class extends Task {

		configure(config) {
		}

		execute(done) {
			var args = taskArray.slice();
			args.push(done);
			this.run.apply(this, args);
		}
	}
}

//登録済みのタスク
var tasks = {};

//タスクの検索
module.exports = function (pathname) {
	//ベースディレクトリを補完
	var base = util.getBase();
	if (base) {
		pathname = path.join(base, pathname);
	}

	//実処理呼び出し
	findTasks(pathname);
};

//タスクを検索して登録
function findTasks(pathname, base) {
	//_で始まるファイルは無視
	if (path.basename(pathname).startsWith("_")) {
		return;
	}

	//存在チェック
	var stat = fs.statSync(pathname);

	//ディレクトリの場合
	if (stat.isDirectory()) {
		//ベース名がなければそのパス
		if (!base) {
			base = pathname;
		}

		//子要素を検索して再帰
		for (let name of fs.readdirSync(pathname)) {
			findTasks(path.join(pathname, name), base);
		}
		return;
	} else {
		//JSファイル以外は無視
		if (path.extname(pathname) != ".js") {
			return;
		}

		//ベースがなければディレクトリ名を使用		
		if (!base) {
			base = path.dirname(pathname);
		}

		//パスの正規化
		pathname = path.normalize(pathname);
		base = path.normalize(base);

		//ベースを起点としたディレクトリ名を取得
		var subname = path.dirname(pathname).substr(base.length + 1);

		//パス階層を生成(fooを起点に/bar/task.jsがあればbar:taskとする)
		var taskpath = [];
		if (subname.length > 0) {
			for (let name of subname.split(path.sep)) {
				taskpath[taskpath.length] = name;
			}
		}
		taskpath[taskpath.length] = path.basename(pathname, ".js");

		//セミコロンでつなげてタスク名にする
		var taskname = taskpath.join(":");

		//gulpに登録
		gulp.task(taskname, function (done) {
			//タスクを取得
			var task = tasks[taskname];

			//なければインスタンス生成
			if (!task) {
				//クラスを取得
				let TaskClass = loader.loadObject(pathname);

				//クラスでなければクラスにする
				if (!isTaskSubClass(TaskClass)) {
					TaskClass = createTaskClass(TaskClass);
				}

				//インスタンスを生成
				task = new TaskClass();
				tasks[taskname] = task;

				//セットアップ
				task.setup(taskname, config.get(taskpath));
			}

			//タスクの実行
			return task._executeInternal(done);
		});
	}
};
