/*!
 * gulpis
 *
 * Copyright (c) 2016 Katsunori Koyanagi.
 * Licensed under the MIT license.
 */

"use strict";

var gulp = require("gulp");
var gulpUtil = require("gulp-util");
var options = require("./options");
var env = require("./env");
var loader = require("./loader");
var config = require("./config");
var util = require("./util");
var version = require("./version");
var cli = require("./cli");

gulp.setMaxListeners(1000);

var moduleBase = (function () {
	var moduleNames = "assert,buffer,child_process,cluster,console," +
		"crypto,dgram,dns,domain,events,fs,http,https,net," +
		"os,path,process,punycode,querystring,readline,stream," +
		"string_decoder,tls,tty,url,util,v8,vm";
	var result = {};

	for (let moduleName of moduleNames.split(",")) {
		result[moduleName] = require(moduleName);
	}

	return result;
})();

//監視ヘルパー
class Watcher {

	//コンストラクタ
	constructor(task) {
		this._task = task;
		this._globs = [];
		this._options = {};
	}

	//globを追加
	when() {
		this._globs = this._globs.concat(Array.prototype.slice.call(arguments));
		return this;
	}

	//オプションを追加
	options(options) {
		for (let key in options) {
			this._options[key] = options[key];
		}
		return this;
	}

	//タスクもしくはコールバックを実行
	then(taskOrCallback) {
		var task;
		if (util.isString(taskOrCallback)) {
			task = _ => this._task.run(taskOrCallback);
		} else {
			task = taskOrCallback;
		}

		this._task.watch(this._globs, this._options, task);
	}
}

//グローバル変数
var context = {};

module.exports = class Task {

	//コンストラクタ
	constructor() {
		this._pluginRequests = [];
		this._moduleRequests = [];
		this._pluginRequested = {};
		this._moduleRequested = Object.create(moduleBase);
		this._plugins = {};
		this._modules = Object.create(moduleBase);
		this._config = {};
		this._name = "";
		this._lastStream = null;
	}

	//最後のストリーム
	get last() {
		return this._lastStream;
	}

	//ワーキングディレクトリ
	get cwd() {
		return util.cwd;
	}

	//タスク間の共有の値
	get context() {
		return context;
	}

	//初期化
	configure(config) {
	}

	//セットアップ
	setup(name, config) {
		//名前と設定
		this._name = name;
		this._config = config;

		//初期化
		this.configure(config);

		//プラグインをロード
		loader.loadModuleSet(this._pluginRequests, this._plugins);
		loader.loadModuleSet(this._moduleRequests, this._modules);

		//ハイフン区切りのモジュールをキャメルケースに変換
		for (let key in this._modules) {
			let key2 = key.replace(/-[a-z]/g, m => m[1].toUpperCase());
			if (key != key2 && !(key2 in this._modules)) {
				this._modules[key2] = this._modules[key];
			}
		}

		//終了監視
		gulp.on("task_stop", e => {
			if (e.task == this._name) {
				this.onTaskEnd();
			}
		});
	}

	//内部実行
	_executeInternal(done) {
		this.onTaskStart();
		return this.execute(done);
	}

	//プラグインを登録
	plugin(module, alias) {
		if (module.startsWith("gulp-")) {
			module = module.substr(5);
		}
		let info = loader.getModuleInfo(module);
		let prop = alias || info.name;
		info.name = "gulp-" + info.name;

		if (info.name in this._pluginRequested) {
			return;
		}
		this._pluginRequested[info.name] = true;

		this._pluginRequests.push({
			module: info,
			load: true,
			key: prop,
		});
	}

	//モジュールを使用する
	module(module, load) {
		let info = loader.getModuleInfo(module);

		if (info.name in this._moduleRequested) {
			return;
		}
		this._moduleRequested[info.name] = true;

		this._moduleRequests.push({
			module: info,
			load: load,
			key: info.name,
		});
	}

	//ストリームにアタッチする
	_attachStream(stream) {
		for (let prop in this._plugins) {
			if (prop in stream) {
				continue;
			}
			this._attachStream0(stream, prop);
		}

		if (!("dest" in stream)) {
			stream.dest = function () {
				return stream.pipe(gulp.dest.apply(gulp, arguments));
			}
		}

		var origin = stream.pipe;

		var self = this;
		stream.pipe = function () {
			var result = origin.apply(stream, arguments);
			self._attachStream(result);
			return result;
		};

		this._lastStream = stream;
	}

	//ストリームにアタッチする
	_attachStream0(stream, prop) {
		var plugin = this._plugins[prop];

		var self = this;
		stream[prop] = function () {
			var newStream = stream.pipe(plugin.apply(null, arguments));
			self._attachStream(newStream);

			return newStream;
		};
	}

	//ソースを返す
	src() {
		var stream = gulp.src.apply(gulp, arguments);
		this._attachStream(stream);

		return stream;
	}

	//プラグインを返す
	get plugins() {
		return this._plugins;
	}

	//モジュールを返す
	get modules() {
		return this._modules;
	}

	//開発モードであるか
	get development() {
		return this.env.toLowerCase() == "development";
	}

	//製品モードであるか
	get production() {
		return this.env.toLowerCase() == "production";
	}

	//ステージングモードであるか
	get staging() {
		return this.env.toLowerCase() == "staging";
	}

	//指定のタスクの設定を返す
	getConfig(name) {
		return config.get(name.split(":"));
	}

	//gulpを返す
	get gulp() {
		return gulp;
	}

	//再起動する
	restart() {
		util.restart();
	}

	//環境変数をSTAGINGに切り替える
	switchToStaging() {
		util.restart("staging");
	}

	//環境変数をPRODUCTIONに切り替える
	switchToProduction() {
		util.restart("production");
	}

	//環境変数をDEVELOPMENTに切り替える
	switchToDevelopment() {
		util.restart("development");
	}

	//プロジェクトテンプレートJSONからプロジェクトを構築
	initProjectFromTemplate(template, distBaseDir, srcBaseDir) {
		util.initProjectFromTemplate(template, distBaseDir, srcBaseDir);
	}

	//CLIを開始
	startCLI(commands) {
		cli(commands, task => this.run(task));
	}

	//実行オプションを返す
	get options() {
		return options;
	}

	//設定を返す
	get config() {
		return this._config;
	}

	//名前を返す
	get name() {
		return this._name;
	}

	//NODE_ENVを返す
	get env() {
		return env;
	}

	//タスク終了
	onTaskEnd() {
	}

	//タスク開始
	onTaskStart() {
	}

	//タスクの実行
	execute(done) {
		done();
	}

	//パスを結合する
	join() {
		return require("path").join.apply(null, arguments);
	}

	//何もしないストリーム生成
	get noop() {
		return gulpUtil.noop;
	}

	//ログ
	log() {
		gulpUtil.log.apply(null, arguments);
	}

	//通知
	notify(title, message) {
		var notifier = loader.loadModule("node-notifire");
		notifier.notify({
			message,
			title,
			wait: false,
		});
	}

	//タスクの実行
	run() {
		var run = loader.loadModule("run-sequence");
		var tasks = Array.prototype.slice.call(arguments);
		var callback = null;

		if (util.isFunction(tasks[tasks.length - 1])) {
			callback = tasks.pop();
		}

		return new Promise((resolve, reject) => {
			tasks.push(_ => {
				if (callback) {
					callback();
				}
				resolve(this.name);
			});
			run.apply(null, tasks);
		});
	}

	//ファイルの監視
	watch() {
		var watch = loader.loadModule("gulp-watch");
		var args = Array.prototype.slice.call(arguments);

		for (var i = 0; i < args.length; i++) {
			if (util.isFunction(args[i])) {
				var fn = args[i];
				args[i] = _ => fn(this.name);
			}
		}
		return watch.apply(null, args);
	}

	//ファイル監視のヘルパを返す
	watcher() {
		return new Watcher(this);
	}

	//バージョン情報
	get version() {
		return version;
	}
};
