/*!
 * gulpis
 *
 * Copyright (c) 2016 Katsunori Koyanagi.
 * Licensed under the MIT license.
 */

"use strict";

var path = require("path");
var fs = require("fs");
var child_process = require("child_process");
var cwd = process.cwd();
var env = require("./env");
var isWindows = process.platform === 'win32';
var started = false;

module.exports = {

	//起動直後のカレントワーキングディレクトリ
	get cwd() {
		return cwd;
	},

	cmd(name) {
		if (isWindows) {
			return name + ".cmd";
		}

		return name;
	},

	//モードに対する終了コード
	envCodes: {
		"": 254,
		"development": 253,
		"staging": 252,
		"production": 251,
	},

	//終了コードに対するモード
	envNames: {
		254: "",
		253: "development",
		252: "staging",
		251: "production",
	},

	//ビルド情報
	about: _ => {
		var gulp = require("gulp");
		var tasks = process.argv.slice(2).filter(arg => arg.indexOf("--") != 0);
		if (tasks.length == 0) {
			tasks = ["default"];
		}

		return `--------------------------------------------------
Gulpis version ${require("../package.json").version}
mode = ${env}
requested tasks = ${tasks.join(", ")}
registered tasks = ${Object.keys(gulp.tasks).join(", ")}
--------------------------------------------------`;
	},

	//開始をスケジュールする
	start() {
		if (started) {
			return;
		}

		started = true;
		setTimeout(_ => require("../bin/cli"), 0);
	},

	//再起動する
	restart(env) {
		env = env || "";
		if (env in this.envCodes) {
			process.exit(this.envCodes[env]);
			return;
		}

		process.exit(envCodes[""]);
	},

	//ベースディレクトリ取得
	getBase() {
		var mod = module;
		while (mod) {
			if (mod.parent && mod.parent.parent == null) {
				return path.dirname(mod.filename);
			}
			mod = mod.parent;
		}

		return "";
	},

	//オブジェクト判定
	isObject(val) {
		if (val == null) {
			return false;
		}

		return Object.prototype.toString.call(val) == "[object Object]";
	},

	//文字列判定
	isString(val) {
		if (val == null) {
			return false;
		}

		return Object.prototype.toString.call(val) == "[object String]";
	},

	//関数判定
	isFunction(val) {
		if (val == null) {
			return false;
		}

		return Object.prototype.toString.call(val) == "[object Function]";
	},

	//配列判定
	isArray(val) {
		if (val == null) {
			return false;
		}

		return Array.isArray(val);
	},

	//プロジェクトテンプレートからプロジェクト作成
	initProjectFromTemplate(template, distBaseDir, srcBaseDir) {
		var srcDir = path.join(this.getBase() || this.cwd, srcBaseDir);
		var distDir = path.join(this.cwd, distBaseDir);
		this.initProjectFromTemplate0(template, distDir, srcDir);
	},

	//プロジェクトテンプレートからプロジェクト作成
	initProjectFromTemplate0(info, distDir, srcDir) {
		//ディレクトリがなければ作成
		try {
			let stat = fs.statSync(distDir);

			//ファイルなら諦める
			if (!stat.isDirectory()) {
				return;
			}
		} catch (e) {
			if (e.code != "ENOENT") {
				throw e;
			}
			fs.mkdirSync(distDir);
			require("gulp-util").log(`Directory created.(${distDir})`);
		}

		//子要素の処理
		for (let key in info) {
			//パス作成
			var pathname = path.join(distDir, key);
			var subInfo = info[key];

			//すでにエントリがあるか
			try {
				let stat = fs.statSync(pathname);

				//ディレクトリではなければ何もしない
				if (!stat.isDirectory()) {
					continue;
				}

				//ディレクトリに対して文字列指定の場合は何もしない
				if (this.isString(subInfo)) {
					continue;
				}

				//オブジェクトでなければ空のオブジェクトにする
				if (!this.isObject(subInfo)) {
					subInfo = {};
				}

				//再帰
				this.initProjectFromTemplate0(subInfo, pathname, srcDir);
				continue;
			} catch (e) {
				if (e.code != "ENOENT") {
					throw e;
				}
			}

			//文字列ならファイルコピー
			if (this.isString(subInfo)) {
				let buf = new Buffer(1);
				let infile = path.join(srcDir, subInfo);
				let input = fs.openSync(infile, "r");
				let output = fs.openSync(pathname, "w");
				let len = 0;

				while (len = fs.readSync(input, buf, 0, buf.length)) {
					fs.writeSync(output, buf, 0, len);
				}

				fs.closeSync(input);
				fs.closeSync(output);
				require("gulp-util").log(`File copied.(${infile} to ${pathname})`);
				continue;
			}

			//オブジェクトでなければ空のオブジェクトにする
			if (!this.isObject(subInfo)) {
				subInfo = {};
			}

			//再帰
			this.initProjectFromTemplate0(subInfo, pathname, srcDir);
		}
	},

	//JSONのマージ
	mergeJson(base, override) {
		//ディープクローン作成
		var base = JSON.parse(JSON.stringify(base));
		var override = JSON.parse(JSON.stringify(override));
		var self = this;

		function mergeJson0(base, override) {
			for (let key of Object.keys(override)) {
				let val = override[key];

				//オブジェクト型でなければ上書き
				if (!self.isObject(val)) {
					base[key] = val;
				} else {
					//メンバーで再帰処理
					let baseVal = base[key];
					if (!self.isObject(baseVal)) {
						baseVal = base[key] = {};
					}

					mergeJson0(baseVal, val);
				}
			}
		}

		//マージ処理呼び出し
		mergeJson0(base, override);

		//マージ結果を返す
		return base;
	},
};
